"""
Config Manager: Runtime Apps Config
-----------------------------------------------
Returns runtime configuration of running server
"""
from typing import Optional
import socket
import os

from hopeit.server.runtime import server

from hopeit.app.context import EventContext
from hopeit.app.logger import app_extra_logger

from hopeit.config_manager import RuntimeAppInfo, RuntimeApps, ServerInfo, ServerStatus
from hopeit.app.api import event_api

logger, extra = app_extra_logger()

__steps__ = ['get_apps_config']

__api__ = event_api(
    summary="Config Manager: Runtime Apps Config",
    description="Returns the runtime config for the Apps running on this server",
    query_args=[("url", Optional[str], "URL used to reach this server, informative")],
    responses={
        200: (RuntimeApps, "Config info about running apps in current process"),
    }
)


async def get_apps_config(payload: None, context: EventContext, *, url: str = "in-process") -> RuntimeApps:
    return RuntimeApps(
        apps={
            app_key: RuntimeAppInfo(
                servers=[ServerInfo(
                    host_name=socket.gethostname(),
                    pid=str(os.getpid()),
                    url=url
                )],
                app_config=app_engine.app_config
            )
            for app_key, app_engine in server.app_engines.items()
        },
        server_status={
            url: ServerStatus.ALIVE
        }
    )



"""
Cluster Apps Config
----------------------------------------------------------------------------------------
Handle remote access to runtime configuration from a group of hosts running hopeit.engine
"""
from hopeit.app.context import EventContext
from hopeit.app.logger import app_extra_logger
from hopeit.app.api import event_api

from hopeit.config_manager import RuntimeApps, client

logger, extra = app_extra_logger()

__steps__ = ['get_hosts_apps_config']

__api__ = event_api(
    summary="Config Manager: Cluster Apps Config",
    description="Handle remote access to runtime configuration for a group of hosts",
    query_args=[("hosts", str, "Comma-separated list of http://host:port strings")],
    responses={
        200: (RuntimeApps, "Combined config info about running apps in provided list of hosts")
    }
)


async def get_hosts_apps_config(payload: None, context: EventContext, *, hosts: str) -> RuntimeApps:
    return await client.get_apps_config(hosts, context)



"""
Config Manager dataclasses
"""
from typing import Dict, List
from enum import Enum
from dataclasses import dataclass, field

from hopeit.dataobjects import dataobject
from hopeit.app.config import AppConfig


@dataobject
@dataclass
class ServerInfo:
    """
    Server info associated with runtime apps
    """
    host_name: str
    pid: str
    url: str = "in-process"


class ServerStatus(Enum):
    ALIVE = "ALIVE"
    ERROR = "ERROR"


@dataobject
@dataclass
class RuntimeAppInfo:
    """
    Application config information associated to servers at runtime
    """
    servers: List[ServerInfo]
    app_config: AppConfig


@dataobject
@dataclass
class RuntimeApps:
    """
    Combined App Config and Server Sttus information for running apps
    """
    apps: Dict[str, RuntimeAppInfo]
    server_status: Dict[str, ServerStatus] = field(default_factory=dict)



"""
Config Manager Client
"""
from typing import Dict, Union, Tuple, Optional
import asyncio
import random

import aiohttp

from hopeit.server.version import APPS_ROUTE_VERSION
from hopeit.app.context import EventContext
from hopeit.server.logger import engine_extra_logger

from hopeit.config_manager import RuntimeApps, RuntimeAppInfo, ServerStatus

logger, extra = engine_extra_logger()


async def get_apps_config(hosts: str, context: Optional[EventContext] = None) -> RuntimeApps:
    """
    Gathers RuntimeApps (runtime apps config) from a given list of hosts running
    `hopeit.config-manager` plugins and returns a combined RuntimeApps
    specifiying for each app_key its configuration and description of hosts where
    it is avalable.

    :param: hosts, str: comma-separated list of the form `http://host:port` where to reach
        servers running hopeit.engine with enabled `config-manager` plugin.
    :return: RuntimeApps, combined from all requested hosts
    """
    responses = await asyncio.gather(
        *[
            _get_host_config(host, context)
            for host in hosts.split(',')
        ]
    )

    apps: Dict[str, RuntimeAppInfo] = {}
    server_status: Dict[str, ServerStatus] = {}
    for host, runtime_apps_response in responses:
        if isinstance(runtime_apps_response, RuntimeApps):
            _combine_apps(apps, runtime_apps_response)
            server_status[host] = runtime_apps_response.server_status.get(host, ServerStatus.ALIVE)
        elif isinstance(runtime_apps_response, ServerStatus):
            server_status[host] = runtime_apps_response

    return RuntimeApps(apps=apps, server_status=server_status)


async def _get_host_config(host: str,
                           context: Optional[EventContext] = None
                           ) -> Tuple[str, Union[RuntimeApps, ServerStatus]]:
    """
    Invokes config-manager runtime-apps-config endpoint in a given host
    """
    # Random <1 sec pause to prevent network overload
    await asyncio.sleep(random.random())

    url = f"{host}/api/config-manager/{APPS_ROUTE_VERSION}/runtime-apps-config?url={host}"
    logger.info(context or __name__, "Invoking config-manager on host: %s...", host, extra=extra(
        host=host, url=url
    ))

    try:
        async with aiohttp.ClientSession() as client:
            async with client.get(url) as response:
                return (host, RuntimeApps.from_dict(await response.json()))  # type: ignore
    except Exception as e:  # pylint: disable=broad-except
        logger.error(context or __name__, "Error contacting host: %s", host, extra=extra(
            host=host, url=url, error=str(e)
        ))
        return (host, ServerStatus.ERROR)


def _combine_apps(apps: Dict[str, RuntimeAppInfo], runtime_apps: RuntimeApps):
    for app_key, app_info in runtime_apps.apps.items():
        app = apps.get(app_key)
        if app is None:
            apps[app_key] = app_info
        else:
            apps[app_key].servers.extend(app_info.servers)
